#!python
import argparse
import re
from pathlib import Path

double_dollars = re.compile(r"\$\$[^$]*\$\$")
simple_dollar = re.compile(r"\$[^$]*\$")


def rewrite(text):
    found = double_dollars.search(text)
    while found:
        sub = text[found.start() + 2 : found.end() - 2]
        sub = sub.strip().replace("+", "+")
        text = (
            text[: found.start()]
            + f'<img src="https://render.githubusercontent.com/render/math?math={sub}">\n\n'
            + text[found.end() :]
        )
        found = double_dollars.search(text)

    found = simple_dollar.search(text)
    while found:
        sub = text[found.start() + 1 : found.end() - 1]
        sub = sub.strip().replace("+", "+")
        text = (
            text[: found.start()]
            + f'<img src="https://render.githubusercontent.com/render/math?math={sub}">'
            + text[found.end() :]
        )
        found = simple_dollar.search(text)
    return text


def parse_files(path, file_type="md"):
    if path.is_dir():
        files = path.glob(f"*.{file_type}")
    else:
        files = [path]
    return files


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "Replace LaTeX expressions (bounded by '$' and '$$')"
            "in all markdown files of a given folder."
        )
    )
    parser.add_argument(
        "-p",
        "--path",
        dest="p",
        type=str,
        required=True,
        help="Path to the folder containing the markdown files, or a markdown itself",
    )
    parser.add_argument(
        "-s",
        "--suffix",
        dest="s",
        type=str,
        default="",
        help="Suffix to add at the end of the name of the modified file.\n If not provided, the files will be modified in place",
    )
    parser.add_argument(
        "-t",
        "--type",
        dest="t",
        type=str,
        default="md",
        choices=["md", "ipynb", "nb"],
        help="Whether to process markdown files directly (option 'md') or notebooks (options 'ipynb' or 'nb')",
    )
    args = parser.parse_args()
    p = Path(args.p)
    if args.t == "nb":
        args.t = "ipynb"
    files = parse_files(p, args.t)
    for file in files:
        if args.t == "ipynb":
            from jupytext import jupytext

            text = jupytext.read(file)
            text = jupytext.notebook_to_md(text)
        else:
            with open(file) as f:
                text = f.read()
        text = rewrite(text)
        if args.s != "" and args.s[0] != ".":
            args.s = "." + args.s
        with open(file.with_suffix(f"{args.s}.md"), "w") as f:
            f.write(text)



# ----------------------------------------------------------------------------
# Title:   Scientific Visualisation - Python & Matplotlib
# Author:  Nicolas P. Rougier
# License: BSD
# ----------------------------------------------------------------------------
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np

n = 100
R, scatter, fig, ax = None, None, None, None


def init_rain():
    global n, R, scatter, fig, ax
    R = np.zeros(
        n,
        dtype=[
            ("position", float, (2,)),
            ("size", float, (1,)),
            ("color", float, (4,)),
        ],
    )
    fig = plt.figure(figsize=(6, 2), facecolor="white", dpi=100)
    ax = fig.add_axes([0, 0, 1, 1], frameon=False)  # , aspect=1)
    scatter = ax.scatter(
        [], [], s=[], linewidth=0.5, edgecolors=[], facecolors="None"
    )


def rain_update(frame):
    global R, scatter

    R["color"][:, 3] = np.maximum(0, R["color"][:, 3] - 1 / len(R))
    R["size"] += 1 / len(R)

    i = frame % len(R)
    R["position"][i] = np.random.uniform(0, 1, 2)
    R["size"][i] = 0
    R["color"][i, 3] = 1

    scatter.set_edgecolors(R["color"])
    scatter.set_sizes(1000 * R["size"].ravel())
    scatter.set_offsets(R["position"])

    return (scatter,)


def run_rain():
    init_rain()
    global R, fig, ax
    R["position"] = np.random.uniform(0, 1, (n, 2))
    R["size"] = np.linspace(0, 1, n).reshape(n, 1)
    R["color"][:, 3] = np.linspace(0, 1, n)

    ax.set_xlim(0, 1), ax.set_xticks([])
    ax.set_ylim(0, 1), ax.set_yticks([])

    anim = animation.FuncAnimation(fig, rain_update, interval=10, frames=200)
    return anim



# ----------------------------------------------------------------------------
# Title:   Scientific Visualisation - Python & Matplotlib
# Author:  Nicolas P. Rougier
# License: BSD
# ----------------------------------------------------------------------------
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import AutoMinorLocator, FuncFormatter, MultipleLocator


def run_anatomy():
    np.random.seed(123)

    X = np.linspace(0.5, 3.5, 100)
    Y1 = 3 + np.cos(X)
    Y2 = 1 + np.cos(1 + X / 0.75) / 2
    Y3 = np.random.uniform(Y1, Y2, len(X))

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(1, 1, 1, aspect=1)

    def minor_tick(x, pos):
        if not x % 1.0:
            return ""
        return "%.2f" % x

    ax.xaxis.set_major_locator(MultipleLocator(1.000))
    ax.xaxis.set_minor_locator(AutoMinorLocator(4))
    ax.yaxis.set_major_locator(MultipleLocator(1.000))
    ax.yaxis.set_minor_locator(AutoMinorLocator(4))
    ax.xaxis.set_minor_formatter(FuncFormatter(minor_tick))

    ax.set_xlim(0, 4)
    ax.set_ylim(0, 4)

    ax.tick_params(which="major", width=1.0)
    ax.tick_params(which="major", length=10)
    ax.tick_params(which="minor", width=1.0, labelsize=10)
    ax.tick_params(which="minor", length=5, labelsize=10, labelcolor="0.25")

    ax.grid(linestyle="--", linewidth=0.5, color=".25", zorder=-10)

    ax.plot(X, Y1, c=(0.25, 0.25, 1.00), lw=2, label="Blue signal", zorder=10)
    ax.plot(X, Y2, c=(1.00, 0.25, 0.25), lw=2, label="Red signal")
    ax.plot(
        X,
        Y3,
        linewidth=0,
        marker="o",
        markerfacecolor="w",
        markeredgecolor="k",
    )

    ax.set_title(
        "Anatomy of a figure", fontsize=20, verticalalignment="bottom"
    )
    ax.set_xlabel("X axis label")
    ax.set_ylabel("Y axis label")

    ax.legend(loc="upper right")

    def circle(x, y, radius=0.15):
        from matplotlib.patches import Circle
        from matplotlib.patheffects import withStroke

        circle = Circle(
            (x, y),
            radius,
            clip_on=False,
            zorder=10,
            linewidth=1,
            edgecolor="black",
            facecolor=(0, 0, 0, 0.0125),
            path_effects=[withStroke(linewidth=5, foreground="w")],
        )
        ax.add_artist(circle)

    def text(x, y, text):
        ax.text(
            x,
            y,
            text,
            backgroundcolor="white",
            # fontname="Yanone Kaffeesatz", fontsize="large",
            ha="center",
            va="top",
            weight="regular",
            color="#000099",
        )

    # Minor tick
    circle(0.50, -0.10)
    text(0.50, -0.32, "Minor tick label")

    # Major tick
    circle(-0.03, 4.00)
    text(0.03, 3.80, "Major tick")

    # Minor tick
    circle(0.00, 3.50)
    text(0.00, 3.30, "Minor tick")

    # Major tick label
    circle(-0.15, 3.00)
    text(-0.15, 2.80, "Major tick label")

    # X Label
    circle(1.80, -0.27)
    text(1.80, -0.45, "X axis label")

    # Y Label
    circle(-0.27, 1.80)
    text(-0.27, 1.6, "Y axis label")

    # Title
    circle(1.60, 4.13)
    text(1.60, 3.93, "Title")

    # Blue plot
    circle(1.75, 2.80)
    text(1.75, 2.60, "Line\n(line plot)")

    # Red plot
    circle(1.20, 0.60)
    text(1.20, 0.40, "Line\n(line plot)")

    # Scatter plot
    circle(3.20, 1.75)
    text(3.20, 1.55, "Markers\n(scatter plot)")

    # Grid
    circle(3.00, 3.00)
    text(3.00, 2.80, "Grid")

    # Legend
    circle(3.70, 3.80)
    text(3.70, 3.60, "Legend")

    # Axes
    circle(0.5, 0.5)
    text(0.5, 0.3, "Axes")

    # Figure
    circle(-0.3, 0.65)
    text(-0.3, 0.45, "Figure")

    color = "#000099"
    ax.annotate(
        "Spines",
        xy=(4.0, 0.35),
        xytext=(3.3, 0.5),
        color=color,
        weight="regular",  # fontsize="large", fontname="Yanone Kaffeesatz",
        arrowprops=dict(arrowstyle="->", connectionstyle="arc3", color=color),
    )

    ax.annotate(
        "",
        xy=(3.15, 0.0),
        xytext=(3.45, 0.45),
        color=color,
        weight="regular",  # fontsize="large", fontname="Yanone Kaffeesatz",
        arrowprops=dict(arrowstyle="->", connectionstyle="arc3", color=color),
    )

    # plt.savefig("../../figures/anatomy/anatomy.pdf")
    plt.show()



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from IPython.display import HTML
from matplotlib import animation
from scipy.ndimage import convolve


def da_alone(a, dt, k):
    """
    Computes the change of concentration given an initial concentration `a`,
    a time increment `dt` and a constant `k`. Note that there is no `i` in
    the equation since it is supposed to be `a` alone.

    Args:
        a (float): The initial concentration value
        dt (float): the time increment
        k (float): the constant k

    Returns:
        (float): the change of concentration
    """
    return dt * (a - a**3 + k)


def da(a, i, dt, k):
    """
    Computes the change of concentration given an initial concentration `a`,
    a time increment `dt` and a constant `k`.

    Args:
        a (float): the initial concentration value
                   of the activator
        i (float): the initial concentration value
                   of the inhibitor
        dt (float): the time increment
        k (float): the constant k

    Returns:
        (float): the change of concentration
    """
    return dt * (a - a**3 - i + k)


def di(i, a, dt, tau):
    """
    Computes the change of concentration given an initial concentration `i`,
    a time increment `dt` and a constant `k`.

    Args:
        a (float): the initial concentration value
                   of the activator
        i (float): the initial concentration value
                   of the inhibitor
        dt (float): the time increment
        tau (float): the constant tau

    Returns:
        (float): the change of concentration
    """
    return dt / tau * (a - i)


def plot_concentration_1cell(c1, c2=None, return_plot=False, save_path=None):
    """
    Plots the concentration evolution given a list of concentrations

    Args:
        c1 (list of floats): a list containing concentrations
        c2 (Optional, list of floats): a list containing concentrations
    """
    fig, ax = plt.subplots(figsize=(5, 4))
    ax.plot(c1, label="First concentration")
    if c2 is not None:
        ax.plot(c2, label="Second concentration")
        ax.legend()
    ax.set_xlabel("Time-point")
    ax.set_ylabel("Concentration")
    fig.tight_layout()
    if save_path is not None:
        save_path = save_path.with_suffix(".png")
        plt.savefig(save_path)
        plt.close(fig)
    if return_plot:
        return fig, ax


def plot_concentration_1D(
    c1, c2=None, return_plot=False, save_path=None, step=100
):
    """
    Plots the concentration evolution given 1 or 2 array(s) of cells
    and concentrations. When two array are combined, the firs concentration
    is in the green and the second in the red. Both values are mixed.

    Args:
        c1 ndarray (n, m): array of concentration of n cells
                           over m time-points
        c2 ndarray (n, m): array of concentration of n cells
                           over m time-points
    """
    fig, ax = plt.subplots(figsize=(10, 8))
    if c2 is None:
        im = ax.imshow(c1[:, ::step], interpolation="nearest")
        cbar = plt.colorbar(im)
        cbar.set_label("Concentration")
    else:
        max_ = np.max([c1, c2])
        min_ = np.min([c1, c2])
        rgb = np.array([c1, c2, np.zeros_like(c1)]).transpose(1, 2, 0)
        rgb = (rgb - min_) / (max_ - min_)
        im = ax.imshow(rgb[:, ::step, :], interpolation="nearest")
    ax.set_xlabel("Time-point")
    ax.set_ylabel("Cell #")
    ax.set_xticks(np.linspace(0, c1.shape[1] // step, 6).astype(int)[:-1])
    ax.set_xticklabels([f"{v*step:.0f}" for v in ax.get_xticks()])
    fig.tight_layout()
    if save_path is not None:
        save_path = save_path.with_suffix(".png")
        plt.savefig(save_path)
        plt.close(fig)
    if return_plot:
        return fig, ax


def __compute_AI(a, i, dt, k, tau, n):
    A, I = [a], [i]
    for t in range(n - 1):
        new_A = A[-1] + da(A[-1], I[-1], dt, k)
        new_I = I[-1] + di(I[-1], A[-1], dt, tau)
        I.append(new_I)
        A.append(new_A)
    return A, I


def retrieve_compute_AI():
    """
    Returns the function compute_AI
    """
    return __compute_AI


def __get_random_table(n, m, seed=0):
    np.random.seed(seed)
    return np.random.rand(n, m)


def get_random_table(n, m, seed=0):
    """
    Return a array of size n by m filled with
    random values between 0 and 1. The random
    values will always be the same thanks to the
    seed. The seed can be changed

    Args:
        n (int): first dimension of the array
        m (int): second dimension of the array
        seed (int): Determine the seed for the
                    random draw. Default 0.
                    If None, it will be different
                    each time it is ran.

    Returns:
        ndarray (n, m)
    """
    # Nope, you will not see the code in here!
    # I mean, if you really want to you can,
    # but that's cheating ...
    return __get_random_table(n, m, seed=seed)


def __question_4(*, A, I, dt, k, tau, n):
    if not isinstance(A, list):
        A = [A]
    if not isinstance(I, list):
        I = [I]
    for t in range(n):
        new_A = A[-1] + da(A[-1], I[-1], dt, k)
        new_I = I[-1] + di(I[-1], A[-1], dt, tau)
        I.append(new_I)
        A.append(new_A)
    return A, I


def __question_13(*, A, I, dt, k, tau, n):
    A = np.copy(A)
    I = np.copy(I)
    for cell_num, (a, i) in enumerate(zip(A[:, 0], I[:, 0])):
        A_cell, I_cell = __compute_AI(a, i, dt, k, tau, n)
        A[cell_num, :] = A_cell
        I[cell_num, :] = I_cell
    return A, I


def __question_14(*, A, I, dt, k, tau, dx, mu_a, mu_i):
    new_A = np.zeros_like(A)
    new_I = np.zeros_like(I)
    new_A[1:-1] = A[1:-1] + dt * (
        dx * mu_a * (A[:-2] + A[2:] - 2 * A[1:-1])
        + A[1:-1]
        - A[1:-1] ** 3
        - I[1:-1]
        + k
    )
    new_A[0] = A[0] + dt * (
        dx * mu_a * (A[1] - A[0]) + A[0] - A[0] ** 3 - I[0] + k
    )
    new_A[-1] = A[-1] + dt * (
        dx * mu_a * (A[-2] - A[-1]) + A[-1] - A[-1] ** 3 - I[-1] + k
    )

    new_I[1:-1] = I[1:-1] + dt / tau * (
        dx * mu_i * (I[:-2] + I[2:] - 2 * I[1:-1]) + A[1:-1] - I[1:-1]
    )
    new_I[0] = I[0] + dt / tau * (dx * mu_i * (I[1] - I[0]) + A[0] - I[0])
    new_I[-1] = I[-1] + dt / tau * (
        dx * mu_i * (I[-2] - I[-1]) + A[-1] - I[-1]
    )
    return new_A, new_I


def __question_16(*, arr, nb_neighbs, kernel, mu, dx, dy):
    to_cell = convolve(arr, kernel, mode="constant", cval=0)
    from_cell = nb_neighbs * arr
    out = mu * (to_cell - from_cell) / (dx * dy)
    return out


def __diffusion(arr, nb_neighbs, kernel, mu, dx, dy):
    to_cell = convolve(arr, kernel, mode="constant", cval=0)
    from_cell = nb_neighbs * arr
    out = mu * (to_cell - from_cell) / (dx * dy)
    return out


def __question_17(*, dt, k, tau, size, T, mu_a, mu_i, dx, dy, seed=0):
    n = int(T / dt)
    A = np.zeros((size, size, n))
    I = np.zeros((size, size, n))
    np.random.seed(seed)
    A[..., 0] = np.random.random((size, size))
    np.random.seed(seed + 1)
    I[..., 0] = np.random.random((size, size))

    kernel = [[0, 1, 0], [1, 0, 1], [0, 1, 0]]
    mask = np.ones_like(A[..., 0])
    nb_neighbs = convolve(mask, kernel, mode="constant", cval=0)

    for t in range(1, n):
        diff_A = __diffusion(A[:, :, t - 1], nb_neighbs, kernel, mu_a, dx, dy)
        A[..., t] = A[..., t - 1] + dt * (
            diff_A + A[..., t - 1] - A[..., t - 1] ** 3 - I[..., t - 1] + k
        )
        diff_I = __diffusion(I[:, :, t - 1], nb_neighbs, kernel, mu_i, dx, dy)
        I[..., t] = I[..., t - 1] + dt / tau * (
            diff_I + A[..., t - 1] - I[..., t - 1]
        )

    return A, I


results_dict = {
    4: __question_4,
    13: __question_13,
    14: __question_14,
    16: __question_16,
    17: __question_17,
}


params_dict = {
    4: """
    For this function, the following calling
    is expected (changing val as needed):
    answer_results(4, A=[<val>], I=[<val>],
                   dt=<val>, k=<val>, tau=<val>)
    """,
    13: """
    For this function, the following calling
    is expected (changing val as needed):
    answer_results(13, A=A, I=I,
                   dt=<val>, k=<val>, tau=<val>)
    With A and I your tables with the first value
    initialized.
    """,
    14: """
    For this function, the following calling
    is expected (changing val as needed):
    answer_results(14, A=A, I=I,
                   dt=<val>, k=<val>, tau=<val>,
                   dx=<val>, mu_a=<val>, mu_i=<val>)
    With A and I a table of size nb_cells*1.
    """,
    16: """
    For this function, the following calling
    is expected (changing val as needed):
    answer_results(16, arr=arr, nb_neighbs=nb_neighbs,
                   kernel=kernel, mu=<val>, dx=<val>, dy=<val>)
    With arr and nb_neighbs a table of size `size`*`size`
    and kernel a table of `size` `s`*`s` with `s<=size`.
    """,
    17: """
    For this function, the following calling
    is expected (changing val as needed):
    answer_results(17, dt=<val>, k=<val>, tau=<val>,
                   size=<val>, T=<val>, mu_a=<val>,
                   mu_i=<val>, dx=<val>, dy=<val>, seed=0)
    with seed being a seed for the random generation
    of the initial concentrations
    """,
}


def answer_results(q, **kwargs):
    """
    Returns the expected output for the question `q`

    Args:
        q (int): the number of the question
        kwargs: the potential args of question q

    Returns:
        ??? It depends on which question was asked
    """
    if q in results_dict:
        try:
            out = results_dict[q](**kwargs)
        except Exception as e:
            print(e)
            print(
                params_dict.get(q, "Unfortunately, no more help is there :/")
            )
            out = None
    else:
        print(f"Question {q} were not found")
        out = None
    return out


from scipy.ndimage import gaussian_filter1d


def __build_curve(
    length=500,
    freq=0.03,
    min_height=1,
    max_height=10,
    start=15,
    start_freq=0.25,
    sigma=3,
):
    """
    Function that builds a curve with randomly positioned spikes.

    Params:
        length (int): length of the curve to create
        freq (float): floating number between 0 and 1.
            It is the frequency at which spikes happen
        min_height (float): minimum height of the spikes
        max_height (float): maximum height of the spikes
        start (float): Percentage at which the spikes start (delay)
        start_freq (float): frequency at wich the traces will
            actually start with a delay
        sigma (float): for nicer rendering, the curve is smoothed
            by a gaussian with the provided sigma

    Returns:
        curve (np.array): a 1d ndarray
    """
    spikes = np.random.rand(length) < freq
    start = start if start_freq < np.random.rand() else 0
    if start != 0:
        spikes[: int(length * start / 100) + sigma] = 0
        spikes[int(length * start / 100) + sigma] = True
    height = min_height + np.random.rand(length) * (max_height - min_height)
    curve = np.zeros_like(spikes, dtype=float)
    curve[spikes] = height[spikes]
    curve = gaussian_filter1d(curve, sigma=sigma)
    return curve


def __solution_1():
    nb_plots = 50
    max_height = 5
    min_height = 1
    length = 500
    freq = 0.03
    sigma = 3
    start = 25
    start_freq = 0.25
    cmap = mpl.cm.get_cmap("Spectral")
    fig, ax = plt.subplots(figsize=(10, 8))
    X = np.linspace(0, 10, length)
    for i in range(nb_plots):
        Y = build_curve(
            length, freq, min_height, max_height, start, start_freq, sigma
        )
        ax.plot(X, 3 * Y + i, color="k", linewidth=5, zorder=nb_plots - i)
        ax.plot(X, [i] * len(Y), color="k", linewidth=5, zorder=nb_plots - i)
        color = cmap(i / nb_plots)
        ax.fill_between(X, 3 * Y + i, i, color=color, zorder=nb_plots - i)

    ax.yaxis.set_tick_params(tick1On=False)
    ax.set_xlim(0, 10)
    ax.set_ylim(-1, nb_plots + max_height)
    ax.axvline(10 * start / 100, ls="--", lw=0.75, color="black", zorder=250)
    ax.set_xticks(range(1, 10, 2))
    ax.spines["bottom"].set_bounds(1, 9)
    ax.yaxis.set_tick_params(labelleft=True)
    ax.set_yticks(np.arange(nb_plots))
    ax.set_yticklabels([f"Trace {i:02d}" for i in range(1, nb_plots + 1)])
    ax.set_xlabel(r"Time [$min$] (fake)")
    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize(10)
        tick.label.set_verticalalignment("bottom")

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_visible(False)
    # Here is the offset
    ax.spines["bottom"].set_position(("outward", 2))

    fig.tight_layout()
    fig.savefig("Resources/exercice_1.png")


def __solution_2(data1):
    fig, ax = plt.subplots(figsize=(6, 5))
    for bin_num in range(10, 52, 20):
        ax.hist(
            data1,
            bins=bin_num,
            histtype="step",
            cumulative=True,
            lw=3,
            label=f"{bin_num} bins",
            alpha=0.75,
        )
    Y = np.ones_like(data1).cumsum()
    X = np.sort(data1)
    ax.plot(X, Y, lw=3, label="'inf' bins", alpha=0.75)
    ax.set_xlim(np.min(data1), np.max(data1))
    ax.set_ylim(0, len(data1) + 1)
    ax.legend(loc="lower right")
    fig.tight_layout()
    fig.savefig("Resources/exercise_2.png")


def __solution_3():
    X1 = np.linspace(0, 2 * np.pi)
    X2 = np.linspace(0, 2 * np.pi)
    Y1 = np.sin(X1 * 2) / 2
    Y2 = np.cos(X2)

    def update(t):
        l1.set_data(X1[:t], Y1[:t])
        l2.set_data(X2[:t], Y2[:t])
        if 5 <= t < 30:
            ax.collections.clear()
            ax.fill_between(X1[5:t], Y1[5:t], Y2[5:t], color="y")

    fig, ax = plt.subplots(figsize=(10, 3))
    (l1,) = ax.plot(X1, Y1, "-o", label="sin", lw=5, ms=10, markevery=[-1])
    (l2,) = ax.plot(X2, Y2, "-o", label="cosin", lw=3, ms=10, markevery=[-1])
    ax.legend(loc="lower left")
    sns.despine(trim=True, offset=15, ax=ax)
    fig.tight_layout()
    anim = animation.FuncAnimation(fig, update, frames=50, interval=50)
    HTML(anim.to_jshtml())
    anim.save("Resources/cos-sin.gif")


def __solution_4():
    np.random.seed(1)
    nb = 200
    positions = np.random.uniform(0, 10, size=(nb, 2))

    def update(t):
        global scatter
        scatter.set_facecolors(colors[: t + 1])
        scatter.set_offsets(positions[:t, ...])
        scatter.set_sizes(sizes[:t])
        return (scatter,)

    cmap = mpl.cm.get_cmap("viridis")
    colors = np.linspace(0, 1, nb)
    sizes = np.linspace(10, 600, nb)
    np.random.shuffle(sizes)
    np.random.shuffle(colors)
    colors = cmap(colors)
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.set_aspect("equal")
    scatter = ax.scatter([], [], edgecolor="k", alpha=0.6)
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 11)
    ax.axis("off")
    fig.tight_layout()
    anim = animation.FuncAnimation(fig, update, frames=nb, interval=15)
    HTML(anim.to_jshtml())
    anim.save("Resources/square.gif")


def __solution_5():
    np.random.seed(1)
    nb = 150
    found = 0
    positions = []
    while found < nb:
        next_pos = np.random.uniform(1, 9, size=2)
        if np.sum((next_pos - [5, 5]) ** 2) < 4**2:
            positions.append(next_pos)
            found += 1
    positions = np.array(positions)
    # positions = np.random.uniform(1, 9, size=(nb, 2))

    # positions = positions[np.sum((positions - [5, 5])**2, axis=1)<16]

    def update(t):
        global scatter
        scatter.set_facecolors(colors[: t + 1])
        scatter.set_offsets(positions[:t, ...])
        scatter.set_sizes(sizes[:t])
        return (scatter,)

    cmap = mpl.cm.get_cmap("viridis")
    colors = np.linspace(0, 1, nb)
    sizes = np.linspace(10, 600, nb)
    np.random.shuffle(sizes)
    np.random.shuffle(colors)
    colors = cmap(colors)
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.set_aspect("equal")
    scatter = ax.scatter([], [], edgecolor="k", alpha=0.6)
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    ax.set_xticks([])
    ax.set_yticks([])
    # ax.axis('off')
    fig.tight_layout()
    anim = animation.FuncAnimation(fig, update, frames=nb, interval=25)
    HTML(anim.to_jshtml())
    anim.save("Resources/circle.gif")


def build_curve(
    length=500,
    freq=0.03,
    min_height=1,
    max_height=10,
    start=15,
    start_freq=0.25,
    sigma=3,
):
    """
    Function that builds a curve with randomly positioned spikes.

    Params:
        length (int): length of the curve to create
        freq (float): floating number between 0 and 1.
            It is the frequency at which spikes happen
        min_height (float): minimum height of the spikes
        max_height (float): maximum height of the spikes
        start (float): Percentage at which the spikes start (delay)
        start_freq (float): frequency at wich the traces will
            actually start with a delay
        sigma (float): for nicer rendering, the curve is smoothed
            by a gaussian with the provided sigma

    Returns:
        curve (np.array): a 1d ndarray
    """
    return __build_curve(
        length=length,
        freq=freq,
        min_height=min_height,
        max_height=max_height,
        start=start,
        start_freq=start_freq,
        sigma=sigma,
    )



answer_dict = {
    1: """
mu_a = 2.8e-4
mu_i = 5e-3
tau = .1
k = -.005
size = 100
dx = dy = 2. / size
T = 9.0
dt = .001
n = int(T / dt)""",
    2: """
c = a
a = b
b = c
OR
a, b = b, a""",
    3: """
for t in range(n):
    a = a + da_alone(a, dt, k)""",
    4: """
for t in range(n):
    old_a = A[-1]
    new_a = old_a + da_alone(old_a, dt, k)
    A.append(new_a)
OR
for t in range(n):
    A.append(A[-1] + da_alone(A[-1], dt, k))""",
    5: """
for t in range(n):
    new_A = A[-1] + da(A[-1], I[-1], dt, k)
    new_I = I[-1] + di(I[-1], A[-1], dt, tau)
    I.append(new_I)
    A.append(new_A)

Note that in that case it is necessary
to make a temporary save of the concentration
values, maybe keep that in mind, this might
become handy later :| """,
    6: """
def compute_AI(a, i, dt, k, tau, n): # Don't forget to add arguments
    A, I = [a], [i]
    for t in range(n-1): # the -1 is because the first value
                         # is already in the array
        new_A = A[-1] + da(A[-1], I[-1], dt, k)
        new_I = I[-1] + di(I[-1], A[-1], dt, tau)
        I.append(new_I)
        A.append(new_A)
    return A, I""",
    7: """
for test_tau in np.linspace(.05, 1, 5):
    for test_k in np.linspace(-1, 1, 5):
        A, I = answer_results(4, A=0.4, I=0.15, dt=dt, k=test_k, tau=test_tau, n=n)
        plot_concentration_1cell(A, I,
                                 save_path=folder / f'k{test_k}_tau{test_tau}.png')

Note that this answer uses the answer_results function
but you should use your own!""",
    8: """
for file in p.iterdir(): f
    name = file.name  # get the name of the of the file
    if 'png' in name: # check if the file is a png file
                      # Note that it is always safer to
                      # check that you are actually working
                      # with the intended files to avoid
                      # removing unintended files!
        k_val, tau_val_png = name.split('_') # That is assuming that
                                             # your files have the format
                                             # described earlier
        new_folder = p / k_val # k_val is the name of the new folder
                               # p is the path to the folder containing
                               # the new folder, hence new_folder is the
                               # path to the new folder.
        if not new_folder.exists(): # checking if the new folder already exists
            Path.mkdir(new_folder)  # Create it if it does not
        file.rename(new_folder / name) #
        # Path.rename(f, new_folder / name) # Equivalent to the previous line""",
    9: """
A = np.zeros((size, n))""",
    10: """
out = A[:3, -8:]""",
    11: """
A[:, 0] = np.random.random(100)""",
    12: """
I = np.zeros((size, n))
I[:, 0] = np.random.random(100)""",
    13: """
# Compressed solution:
for cell_num, (a, i) in enumerate(zip(A[:, 0], I[:, 0])):
    A_cell, I_cell = compute_AI(a, i, dt, k, tau, n)
    A[cell_num, :] = A_cell
    I[cell_num, :] = I_cell

# Other solution:
for cell_num in range(size):
    a = A[cell_num, 0]
    i = I[cell_num, 0]
    A_cell, I_cell = compute_AI(a, i, dt, k, tau, n)
    A[cell_num, :] = A_cell
    I[cell_num, :] = I_cell

""",
    14: """
def dA_I(A, I, dt, k, tau, dx, mu_a, mu_i):
    new_A = np.zeros_like(A)
    new_I = np.zeros_like(I)
    # Global solution
    new_A[1:-1] = (A[1:-1] +
                   dt * (mu_a/dx*(A[:-2] + A[2:] - 2*A[1:-1]) +
                         A[1:-1] - A[1:-1]**3 - I[1:-1] + k))

    new_I[1:-1] = (I[1:-1] +
                   dt/tau * (mu_i/dx*(I[:-2] + I[2:] - 2*I[1:-1]) +
                             A[1:-1] - I[1:-1]))

    # Edge case handling: first (0) and last (-1) cells
    new_A[0] = (A[0] +
                dt * (mu_a/dx*(A[1] - A[0]) +
                      A[0] - A[0]**3 - I[0] + k))
    new_A[-1] = (A[-1] +
                 dt * (mu_a/dx*(A[-2] - A[-1]) +
                       A[-1] - A[-1]**3 - I[-1] + k))

    new_I[0] = (I[0] +
                dt/tau * (mu_i/dx*(I[1] - I[0]) +
                          A[0] - I[0]))
    new_I[-1] = (I[-1] +
                 dt/tau * (mu_i/dx*(I[-2] - I[-1]) +
                           A[-1] - I[-1]))
    return new_A, new_I""",
    16: """
def diffusion(arr, nb_neighbs, kernel, mu, dx, dy):
    to_cell = convolve(arr, kernel, mode='constant', cval=0)
    from_cell = nb_neighbs*arr
    out = mu*(to_cell - from_cell)/(dx*dy)
    return out""",
    17: """
def compute_turing(dt, k, tau, size, T,
                   mu_a, mu_i, dx, dy, seed=0):
    n = int(T/dt)
    A = np.zeros((size, size, n))
    I = np.zeros((size, size, n))
    np.random.seed(seed)
    A[:, :, 0] = np.random.random((size, size))
    np.random.seed(seed+1)
    I[:, :, 0] = np.random.random((size, size))

    kernel = [[0, 1, 0],
              [1, 0, 1],
              [0, 1, 0]]
    mask = np.ones_like(A[:, :, 0])
    nb_neighbs = convolve(mask, kernel, mode='constant', cval=0)

    for t in range(1, n):
        diff_A = diffusion(A[:, :, t-1], nb_neighbs, kernel, mu_a, dx, dy)
        A[..., t] = A[..., t-1] + dt*(diff_A + A[..., t-1] - A[..., t-1]**3 - I[..., t-1] + k)
        diff_I = diffusion(I[:, :, t-1], nb_neighbs, kernel, mu_i, dx, dy)
        I[..., t] = I[..., t-1] + dt/tau*(diff_I + A[..., t-1] - I[..., t-1])

    return A, I""",
}

hint_dict = {
    3: """
Don't forget that the function
returns the change, therefore,
that value needs to be added to
the original value to be valid.

Moreover, keep in mind that the
number of iterations that you
would like to do is calculated
and stored in the variable n.

Finally, the function da_dt_alone
can be called the following way:
da_dt_alone(a, dt, k)""",
    4: """
To access the last element of a list L
one can use L[-1].

To add an element to a list L, one can
use the function append:
L.append(<element>)""",
    5: """
It is important to remember that
it is necessary to make a copy
of the latest values of the list
before manipulating them.""",
    6: """
It is important to remember to
define the correct function
parameters and return values""",
    7: """
Remember to use the `?` to call
the help of a function (np.linspace).

Because we want to make two parameters
vary simultaneously, we need to have
a for loop within another one!""",
    8: """
Look at the following cells to see some advanced help""",
}


def answer(q):
    """
    Print the answer of the question `q`

    Args:
        q (int): the number of the question
    """
    print(answer_dict.get(q, f"Question {q} not found"))


def hint(q):
    """
    Print the hint of the question `q`

    Args:
        q (int): the number of the question
    """
    print(
        hint_dict.get(q, f"Unfortunately, there is no hint for question {q}")
    )



__version__ = "2024.1.3"

from .anatomy import run_anatomy
from .Answers import answer, hint
from .rain import run_rain
from .utils import (
    answer_results,
    build_curve,
    da,
    da_alone,
    di,
    get_random_table,
    plot_concentration_1cell,
    plot_concentration_1D,
    retrieve_compute_AI,
)

__all__ = [
    "answer",
    "hint",
    "da",
    "da_alone",
    "di",
    "plot_concentration_1cell",
    "plot_concentration_1D",
    "retrieve_compute_AI",
    "get_random_table",
    "answer_results",
    "build_curve",
    "run_rain",
    "run_anatomy",
]
